import os

from django.contrib.auth.decorators import login_required
from django.core.files.storage import default_storage
from django.http import JsonResponse, HttpResponse
from django.template.loader import render_to_string

from async_downloads.cache import get_collection_key
from async_downloads.settings import DOWNLOAD_TEMPLATE, WS_MODE, cache


@login_required
def ajax_update(request):
    download_keys = cache.get(get_collection_key(request.user), [])
    downloads = []
    in_progress = False
    for i, download_key in enumerate(download_keys):
        dl = cache.get(download_key)
        if not dl:
            continue
        if dl["complete"]:
            dl["url"] = default_storage.url(dl["filepath"])
        else:
            in_progress = True
        if WS_MODE:
            dl["timestamp"] = str(dl["timestamp"])
            dl["download_key"] = download_key
        downloads.append(dl)
    # TODO: split up complete and in progress async_downloads?
    return JsonResponse(
        {
            "html": render_to_string(DOWNLOAD_TEMPLATE, {"downloads": downloads}),
            "downloads": downloads,
            "in_progress": in_progress,
        }
    )


@login_required
def ajax_clear_download(request):
    # TODO: consider just clearing the key without deleting,
    #  so that all deletion is done by one function
    filepath = request.POST.get("filepath")
    directory = os.path.split(filepath)[0]
    download_key = os.path.split(directory)[1]
    cache.delete(download_key)
    default_storage.delete(filepath)
    default_storage.delete(directory)
    return HttpResponse("")



from django.urls import path

from async_downloads import views

app_name = "async_downloads"
urlpatterns = [
    path("ajax_update/", views.ajax_update, name="ajax_update"),
    path("ajax_clear_download/", views.ajax_clear_download, name="ajax_clear_download"),
]



import csv
import logging
import os
import uuid
from datetime import datetime
from tempfile import SpooledTemporaryFile

from django.contrib.auth import get_user_model
from django.core.files import File
from django.core.files.storage import default_storage
from pathvalidate import sanitize_filename

from async_downloads.settings import (
    COLLECTION_KEY_FORMAT,
    IN_MEMORY_MAX_SIZE_BYTES,
    CSS_CLASS,
    PATH_PREFIX,
    TIMEOUT,
    WS_MODE,
    cache,
)
from async_downloads.ws_consumers import ws_init_download, ws_update_download

logger = logging.getLogger(__name__)


def get_username(user):
    username = None
    user_model = get_user_model()
    if isinstance(user, user_model):
        username = user.username
    elif isinstance(user, int):
        username = user_model.objects.get(pk=user).username
    else:
        return user
    return username


def get_collection_key(user):
    return COLLECTION_KEY_FORMAT.format(get_username(user))


def init_download(user, filename, name=None):
    download_key = f"{uuid.uuid4()}"
    filename = sanitize_filename(filename)
    filepath = os.path.join(PATH_PREFIX, download_key, filename)
    # TODO: consider inserting the collection key in here, so that it can be
    #  "touched" (`cache.touch`) to avoid the potential of the collection expiring
    #  before its downloads do
    download = {
        "user": get_username(user),
        "timestamp": datetime.now(),
        "filepath": filepath,
        "name": name or filename,
        "complete": False,
        "errors": "",
        "percentage": 0,
        "css_class": CSS_CLASS,
        "url": "",
    }
    collection_key = get_collection_key(user)
    # TODO: locking mechanism - consider https://pypi.org/project/django-cache-lock/
    # TODO: build the cleanup of expired keys into this?
    #  (since we are already modifying the cache entry)
    download_keys = [download_key] + cache.get(collection_key, [])
    cache.set(collection_key, download_keys, TIMEOUT)
    cache.set(download_key, download, TIMEOUT)
    if WS_MODE:
        ws_init_download(download_key)
    return collection_key, download_key


def save_download(download_key, iterable=None, file=None):
    # file is a BytesIO object
    download = cache.get(download_key)
    if not download:
        return
    if iterable is not None:
        try:
            with SpooledTemporaryFile(
                max_size=IN_MEMORY_MAX_SIZE_BYTES,
                mode="w+",
                newline="",
                encoding="utf-8",
            ) as temp_file:
                writer = csv.writer(temp_file, lineterminator="\n")
                for row in iterable:
                    writer.writerow(row)
                # Force any remaining bytes in the buffer
                temp_file.flush()
                default_storage.save(download["filepath"], temp_file._file.buffer)
        except Exception as e:
            logger.exception(
                "Download failed with type: iterable key: %s name: %s",
                download_key,
                download.get("name"),
            )
            download["errors"] = str(e)
    elif file is not None:
        try:
            default_storage.save(download["filepath"], File(file))
        except Exception as e:
            logger.exception(
                "Download failed with type: File key: %s name: %s",
                download_key,
                download.get("name"),
            )
            download["errors"] = str(e)
    download["complete"] = True
    download["percentage"] = 100
    cache.set(download_key, download, TIMEOUT)
    if WS_MODE:
        ws_update_download(download_key)


def set_percentage(download_key, percentage):
    download = cache.get(download_key)
    if not download:
        return
    # Cap percentage between 0 and 100 and ensure it is an int
    download["percentage"] = min(max(0, int(percentage)), 100)
    cache.set(download_key, download, TIMEOUT)
    if WS_MODE:
        ws_update_download(download_key)


def update_percentage(download_key, total, cur, resolution=10):
    # Cap between 1 and 100, as there is no point in higher resolution,
    # and lower than 1 is invalid
    resolution = min(max(1, resolution), 100)
    for x in range(1, resolution):
        if cur + 1 == total // resolution * x:
            set_percentage(download_key, (100 / resolution) * x)
            break


def cleanup_collection(collection_key):
    # TODO: clean up files
    #  need to delete the file first, then the directory
    #    default_storage.delete(f for f in default_storage.listdir(os.path.join(PATH_PREFIX, download_key)))
    #    default_storage.delete(os.path.join(PATH_PREFIX, download_key))
    stored_keys = cache.get(collection_key, [])
    active_keys = [key for key in stored_keys if cache.get(key) is not None]
    # Avoid cache write if possible to reduce chance of clobbering
    if len(active_keys) != len(stored_keys):
        cache.set(collection_key, active_keys)
    # cache.set(
    #     collection_key, [key for key in cache.get(collection_key, []) if cache.get(key) is not None]
    # )



from django.conf import settings
from django.core.cache import cache as default_cache

cache = default_cache
DEFAULT_TIMEOUT = 60 * 60 * 24
DEFAULT_DOWNLOAD_TEMPLATE = "async_downloads/downloads.html"
DEFAULT_PATH_PREFIX = "downloads"
DEFAULT_COLLECTION_KEY_FORMAT = "async_downloads/{}"
DEFAULT_IN_MEMORY_MAX_SIZE_BYTES = 5 * 1e6  # Default of 5MB
DEFAULT_WS_MODE = False
DEFAULT_CACHE_NAME = ''
DEFAULT_CSS_CLASS = None
DEFAULT_WS_CHANNEL_NAME = "downloads"

TIMEOUT = getattr(settings, "ASYNC_DOWNLOADS_TIMEOUT", DEFAULT_TIMEOUT)
DOWNLOAD_TEMPLATE = getattr(settings, "ASYNC_DOWNLOADS_DOWNLOAD_TEMPLATE", DEFAULT_DOWNLOAD_TEMPLATE)
PATH_PREFIX = getattr(settings, "ASYNC_DOWNLOADS_PATH_PREFIX", DEFAULT_PATH_PREFIX)
COLLECTION_KEY_FORMAT = getattr(settings, "ASYNC_DOWNLOADS_COLLECTION_KEY_FORMAT", DEFAULT_COLLECTION_KEY_FORMAT)
IN_MEMORY_MAX_SIZE_BYTES = getattr(settings, "ASYNC_DOWNLOADS_IN_MEMORY_MAX_SIZE_BYTES", DEFAULT_IN_MEMORY_MAX_SIZE_BYTES)
WS_MODE = getattr(settings, "ASYNC_DOWNLOADS_WS_MODE", DEFAULT_WS_MODE)
CACHE_NAME = getattr(settings, "ASYNC_DOWNLOADS_CACHE_NAME", DEFAULT_CACHE_NAME)
CSS_CLASS = getattr(settings, "ASYNC_DOWNLOADS_CSS_CLASS", DEFAULT_CSS_CLASS)
WS_CHANNEL_NAME = getattr(settings, "ASYNC_DOWNLOADS_WS_CHANNEL_NAME", DEFAULT_WS_CHANNEL_NAME)

if CACHE_NAME != DEFAULT_CACHE_NAME:
    from django.core.cache import caches
    cache = caches[CACHE_NAME]

# WS DEFINITIONS
EVENT_TYPE_REFRESH_DOWNLOADS = 'refreshDownloads'
EVENT_TYPE_UPDATE_DOWNLOAD = 'updateDownload'
EVENT_TYPE_INIT_DOWNLOAD = 'initDownload'
EVENT_TYPE_REMOVE_DOWNLOAD = 'removeDownload'
EVENT_TYPE_INIT_ALL_DOWNLOADS = 'initAllDownloads'

USE_WS = False
if WS_MODE:
    USE_WS = True



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='django-async-downloads',
    version='0.4.4',
    author='David Vaughan',
    author_email='david.vaughan@quickrelease.co.uk',
    maintainer="Quick Release (Automotive) Ltd.",
    description='Asynchronous downloads scaffolding for Django projects',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/QuickRelease/django-async-downloads.git',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ],
    keywords="django download asynchronous async celery",
    packages=["async_downloads"],
    include_package_data=True,
    install_requires=[
        "Django>=2.0",
        "celery>=4.2.1",
        "pathvalidate>=2.3.0",
    ],
    extras_require={
        "ws": [
            "channels>=3.0",
            "channels_redis>=3.0",
        ],
    },
    license="MIT",
)



import json
import os

from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async
from channels.exceptions import DenyConnection
from channels.generic.websocket import AsyncWebsocketConsumer
from channels.layers import get_channel_layer
from django.contrib.auth import get_user_model
from django.core.files.storage import default_storage
from django.template.loader import render_to_string

import async_downloads.settings as settings
from async_downloads.settings import DOWNLOAD_TEMPLATE, WS_CHANNEL_NAME, cache


def _timestamp_to_str(timestamp):
    return timestamp.astimezone().isoformat()


def ws_init_download(download_key):
    download = cache.get(download_key)
    download["download_key"] = download_key
    download["timestamp"] = _timestamp_to_str(download["timestamp"])
    download["html"] = render_to_string(DOWNLOAD_TEMPLATE, {"downloads": [download]})
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        f"{WS_CHANNEL_NAME}_{download['user']}",
        {
            "type": "init_download",
            "data": {
                "download": download,
                "download_key": download_key,
            },
        },
    )


def ws_update_download(download_key):
    download = cache.get(download_key)
    download["timestamp"] = _timestamp_to_str(download["timestamp"])
    if download["complete"]:
        download["url"] = default_storage.url(download["filepath"])
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        f"{WS_CHANNEL_NAME}_{download['user']}",
        {
            "type": "update_single_download",
            "data": {
                "download": download,
                "download_key": download_key,
            },
        },
    )


class DownloadsConsumer(AsyncWebsocketConsumer):
    async def connect(self):
        user = self.scope["user"]
        if not user.is_authenticated:
            raise DenyConnection
        self.username = user.username
        self.user = await database_sync_to_async(get_user_model().objects.get)(
            username=self.username
        )
        self.user_group_name = f"{WS_CHANNEL_NAME}_{self.username}"
        await self.channel_layer.group_add(self.user_group_name, self.channel_name)
        await self.accept()

    async def disconnect(self, close_code):
        if hasattr(self, "user_group_name"):
            await self.channel_layer.group_discard(
                self.user_group_name, self.channel_name
            )

    async def receive(self, text_data):
        data = json.loads(text_data)
        event = data["eventType"]
        if event == "clearDownload":
            await self.clear_download(data["data"]["filepath"])
        if event == "initAllDownloads":
            await self.init_all_downloads()

    async def init_all_downloads(self):
        from async_downloads.cache import get_collection_key

        download_keys = cache.get(get_collection_key(self.user), [])
        downloads = []
        #  by default they are listed from newest
        for download_key in reversed(download_keys):
            dl = cache.get(download_key)
            if not dl:
                continue
            if dl["complete"]:
                dl["url"] = default_storage.url(dl["filepath"])
            dl["download_key"] = download_key
            dl["timestamp"] = _timestamp_to_str(dl["timestamp"])
            dl["html"] = render_to_string(DOWNLOAD_TEMPLATE, {"downloads": [dl]})
            downloads.append({"download": dl, "download_key": download_key})
        await self.send(
            json.dumps(
                {"eventType": settings.EVENT_TYPE_INIT_ALL_DOWNLOADS, "data": downloads}
            )
        )

    async def remove_single_download(self, data):
        await self.send(
            json.dumps({"eventType": settings.EVENT_TYPE_REMOVE_DOWNLOAD, **data})
        )

    async def update_single_download(self, data):
        await self.send(
            json.dumps({"eventType": settings.EVENT_TYPE_UPDATE_DOWNLOAD, **data})
        )

    async def init_download(self, data):
        await self.send(
            json.dumps({"eventType": settings.EVENT_TYPE_INIT_DOWNLOAD, **data})
        )

    async def clear_download(self, filepath):
        directory = os.path.split(filepath)[0]
        download_key = os.path.split(directory)[1]
        cache.delete(download_key)
        default_storage.delete(filepath)
        default_storage.delete(directory)
        await self.channel_layer.group_send(
            self.user_group_name,
            {"type": "remove_single_download", "data": {"download_key": download_key}},
        )
